"""
Barycenter/circumcenter optimization.
"""
import numpy
import quadpy
import scipy.sparse.linalg
from meshplex import MeshTri

from .helpers import get_new_points_averaged, runner


def fixed_point_uniform(points, cells, *args, boundary_step=None, **kwargs):
    def get_new_points(mesh):
        # "Switching" based on some criterion is not good because this constitutes a
        # discontinuity in the energy functional, possibly leading to loops in the
        # optimization process.

        # Get circumcenters for good cells, barycenters for bad ones.
        ref = mesh.cell_circumcenters.copy()
        # mask = numpy.any(mesh.ce_ratios < -0.0, axis=0)
        alpha = 0.5
        mask = mesh.cell_quality < alpha
        q = mesh.cell_quality
        print(len(ref), sum(mask), numpy.min(q), numpy.average(q), numpy.max(q))
        ref[mask] = mesh.cell_barycenters[mask]

        X = get_new_points_averaged(mesh, ref, mesh.cell_volumes)
        if boundary_step is None:
            # Reset boundary points to their original positions.
            idx = mesh.is_boundary_point
            X[idx] = mesh.points[idx]
        else:
            # Move all boundary points back to the boundary.
            idx = mesh.is_boundary_point
            X[idx] = boundary_step(X[idx].T).T
        return X

    mesh = MeshTri(points, cells)
    runner(
        get_new_points,
        mesh,
        *args,
        **kwargs,
        method_name="Barycenter/circumcenter optimization (BCO), uniform density, fixed-point variant"
    )
    return mesh.points, mesh.cells["points"]


def _energy_uniform_per_point(X, cells):
    """The CPT mesh energy is defined as

        sum_i E_i,
        E_i = 1/(d+1) * sum int_{omega_i} ||x - x_i||^2 rho(x) dx,

    see Chen-Holst. This method gives the E_i and  assumes uniform density, rho(x) = 1.
    """
    dim = 2
    mesh = MeshTri(X, cells)

    star_integrals = numpy.zeros(mesh.points.shape[0])
    # Python loop over the cells... slow!
    for cell, cell_volume in zip(mesh.cells["points"], mesh.cell_volumes):
        for idx in cell:
            xi = mesh.points[idx]
            tri = mesh.points[cell]
            # Take any scheme with order 2
            scheme = quadpy.triangle.dunavant_02()
            val = scheme.integrate(
                lambda x: numpy.einsum("ij,ij->i", x.T - xi, x.T - xi), tri
            )
            star_integrals[idx] += val

    return star_integrals / (dim + 1)


def energy_uniform(X, cells):
    return numpy.sum(_energy_uniform_per_point(X, cells))


def jac_uniform(X, cells):
    """The approximated Jacobian is

      partial_i E = 2/(d+1) (x_i int_{omega_i} rho(x) dx - int_{omega_i} x rho(x) dx)
                  = 2/(d+1) sum_{tau_j in omega_i} (x_i - b_{j, rho}) int_{tau_j} rho,

    see Chen-Holst. This method here assumes uniform density, rho(x) = 1, such that

      partial_i E = 2/(d+1) sum_{tau_j in omega_i} (x_i - b_j) |tau_j|

    with b_j being the ordinary barycenter.
    """
    dim = 2
    mesh = MeshTri(X, cells)

    jac = numpy.zeros(X.shape)
    for k in range(mesh.cells["points"].shape[1]):
        i = mesh.cells["points"][:, k]
        vals = (mesh.points[i] - mesh.cell_barycenters).T * mesh.cell_volumes
        # numpy.add.at(jac, i, vals)
        jac += numpy.array(
            [numpy.bincount(i, val, minlength=jac.shape[0]) for val in vals]
        ).T

    return 2 / (dim + 1) * jac


def solve_hessian_approx_uniform(X, cells, rhs):
    """As discussed above, the approximated Jacobian is

      partial_i E = 2/(d+1) sum_{tau_j in omega_i} (x_i - b_j) |tau_j|.

    To get the Hessian, we have to form its derivative. As a simplifications,
    let us assume again that |tau_j| is independent of the point positions. Then we get

       partial_ii E = 2/(d+1) |omega_i| - 2/(d+1)**2 |omega_i|,
       partial_ij E = -2/(d+1)**2 |tau_j|.

    The terms with (d+1)**2 are from the barycenter in `partial_i E`. It turns out from
    numerical experiments that the negative term in `partial_ii E` is detrimental to the
    convergence. Hence, this approximated Hessian solver only considers the off-diagonal
    contributions from the barycentric terms.
    """
    dim = 2
    mesh = MeshTri(X, cells)

    # Create matrix in IJV format
    row_idx = []
    col_idx = []
    val = []

    cells = mesh.cells["points"].T
    n = X.shape[0]

    # Main diagonal, 2/(d+1) |omega_i| x_i
    a = mesh.cell_volumes * (2 / (dim + 1))
    for i in [0, 1, 2]:
        row_idx += [cells[i]]
        col_idx += [cells[i]]
        val += [a]

    # terms corresponding to -2/(d+1) * b_j |tau_j|
    a = mesh.cell_volumes * (2 / (dim + 1) ** 2)
    for i in [[0, 1, 2], [1, 2, 0], [2, 0, 1]]:
        edges = cells[i]
        # Leads to funny osciilatory movements
        # row_idx += [edges[0], edges[0], edges[0]]
        # col_idx += [edges[0], edges[1], edges[2]]
        # val += [-a, -a, -a]
        # Best so far
        row_idx += [edges[0], edges[0]]
        col_idx += [edges[1], edges[2]]
        val += [-a, -a]

    row_idx = numpy.concatenate(row_idx)
    col_idx = numpy.concatenate(col_idx)
    val = numpy.concatenate(val)

    # Set Dirichlet conditions on the boundary
    matrix = scipy.sparse.coo_matrix((val, (row_idx, col_idx)), shape=(n, n))
    # Transform to CSR format for efficiency
    matrix = matrix.tocsr()

    # Apply Dirichlet conditions.
    # Set all Dirichlet rows to 0.
    for i in numpy.where(mesh.is_boundary_point)[0]:
        matrix.data[matrix.indptr[i] : matrix.indptr[i + 1]] = 0.0
    # Set the diagonal and RHS.
    d = matrix.diagonal()
    d[mesh.is_boundary_point] = 1.0
    matrix.setdiag(d)

    rhs[mesh.is_boundary_point] = 0.0

    out = scipy.sparse.linalg.spsolve(matrix, rhs)

    # PyAMG fails on circleci.
    # ml = pyamg.ruge_stuben_solver(matrix)
    # # Keep an eye on multiple rhs-solves in pyamg,
    # # <https://github.com/pyamg/pyamg/issues/215>.
    # tol = 1.0e-10
    # out = numpy.column_stack(
    #     [ml.solve(rhs[:, 0], tol=tol), ml.solve(rhs[:, 1], tol=tol)]
    # )
    return out


def quasi_newton_uniform(points, cells, *args, **kwargs):
    """Like linear_solve above, but assuming rho==1. Note that the energy gradient

        \\partial E_i = 2/(d+1) sum_{tau_j in omega_i} (x_i - b_j) \\int_{tau_j} rho

    becomes

        \\partial E_i = 2/(d+1) sum_{tau_j in omega_i} (x_i - b_j) |tau_j|.

    Because of the dependence of |tau_j| on the point coordinates, this is a nonlinear
    problem.

    This method makes the simplifying assumption that |tau_j| does in fact _not_ depend
    on the point coordinates. With this, one still only needs to solve a linear system.
    """

    def get_new_points(mesh):
        # do one Newton step
        # TODO need copy?
        x = mesh.points.copy()
        cells = mesh.cells["points"]
        jac_x = jac_uniform(x, cells)
        x -= solve_hessian_approx_uniform(x, cells, jac_x)
        return x

    mesh = MeshTri(points, cells)
    runner(
        get_new_points,
        mesh,
        *args,
        **kwargs,
        method_name="Centroidal Patch Tesselation (CPT), uniform density, quasi-Newton variant"
    )
    return mesh.points, mesh.cells["points"]
